from src.lpsim.server.event_handler import OmnipotentGuideEventHandler
from src.lpsim.network import HTTPServer
from src.lpsim.server.match import MatchConfig, Match
from tests.utils_for_test import get_random_state
import logging


logging.basicConfig(level = logging.INFO)


if __name__ == '__main__':
    deck_str_1 = '''
        default_version:4.1
        charactor:Noelle
        charactor:Mona
        charactor:Dehya
        # Gales of Reverie*10
        # Beneficent*10
        Stalwart and True*10
        Sweet Madame*10
    '''
    deck_str_2 = '''
    charactor:Nahida@3.7
    charactor:Rhodeia of Loch@3.3
    charactor:Fischl@3.3
    Gambler's Earrings@3.8
    Paimon@3.3
    Chef Mao@4.1
    Dunyarzad@4.1
    The Bestest Travel Companion!@3.3
    Paimon@3.3
    Liben@3.3
    Liben@3.3
    Send Off@3.7
    Teyvat Fried Egg@4.1
    Dunyarzad@4.1
    Sweet Madame@3.3
    I Haven't Lost Yet!@4.0
    Send Off@3.7
    Lotus Flower Crisp@3.3
    Magic Guide@3.3*15
    '''
    server = HTTPServer(
        decks = [deck_str_1, deck_str_1],
        match_config = MatchConfig(
            check_deck_restriction = False,
            card_number = None,
            max_same_card_number = None,
            charactor_number = None,
            max_round_number = 999,
            random_first_player = False,
            # max_hand_size = 999,
            # recreate_mode = True,
            history_level = 10,
            # make_skill_prediction = True,
            # random_object_information = {
            #     'rhodeia': [
            #         'squirrel', 'raptor', 'frog', 'squirrel', 'raptor', 
            #         'frog', 'frog', 'squirrel', 'squirrel'
            #     ],
            # }
        )
    )

    # fix random seed
    match = Match(random_state = get_random_state())
    match.set_deck(server.decks)
    match.config = server.match.config

    # rich mode
    match.event_handlers.append(OmnipotentGuideEventHandler())
    match.config.initial_dice_number = 16

    server.match = match
    server.match.start()
    server.match._save_history()
    server.match.step()
    server.run()



from typing import Any, List, Literal

from ....event import (
    ChooseCharactorEventArguments, SwitchCharactorEventArguments
)

from ....action import CreateObjectAction
from ....modifiable_values import CostValue, DamageIncreaseValue
from ....consts import CostLabels, ObjectPositionType, SkillType
from ....struct import Cost
from .base import RoundEffectArtifactBase


class SkillCostDecreaseArtifact(RoundEffectArtifactBase):
    name: str
    desc: str
    version: str
    cost: Cost
    skill_label: int  # CostLabels

    max_usage_per_round: int = 1

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal['TEST', 'REAL'],
    ) -> CostValue:
        if self.usage > 0:  
            # has usage
            if not self.position.check_position_valid(
                value.position, match,
                player_idx_same = True, 
                source_area = ObjectPositionType.CHARACTOR,
            ):
                # not from self position or not equipped
                return value
            label = value.cost.label
            target = CostLabels.TALENT.value | self.skill_label
            if label & target == 0:
                # no label match
                return value
            position = value.position
            assert self.position.charactor_idx != -1
            if position.area == ObjectPositionType.SKILL:
                # cost from charactor
                if position.charactor_idx != self.position.charactor_idx:
                    # not same charactor
                    return value
            else:
                assert position.area == ObjectPositionType.HAND
                # cost from hand card, is a talent card
                equipped_charactor = match.player_tables[
                    self.position.player_idx
                ].charactors[self.position.charactor_idx]
                for card in match.player_tables[
                        self.position.player_idx].hands:
                    if card.id == value.position.id:
                        if card.charactor_name != equipped_charactor.name:
                            # talent card not for this charactor
                            return value
            # can decrease cost
            if (  # pragma: no branch
                value.cost.decrease_cost(value.cost.elemental_dice_color)
            ):
                # decrease cost success
                if mode == 'REAL':
                    self.usage -= 1
        return value


class ThunderingPoise(SkillCostDecreaseArtifact):
    name: Literal['Thundering Poise']
    desc: str = (
        'When the character uses a Normal Attack or equips a Talent: Spend 1 '
        'less Elemental Die. (Once per Round)'
    )
    version: Literal['4.0'] = '4.0'
    cost: Cost = Cost(any_dice_number = 2)
    skill_label: int = CostLabels.NORMAL_ATTACK.value


class VermillionHereafter(ThunderingPoise):
    name: Literal['Vermillion Hereafter']
    desc: str = (
        'When the character uses a Normal Attack or equips a Talent: Spend 1 '
        'less Elemental Die. (Once per Round) '
        'After a character is switched to the active character: During this '
        'Round, character deals +1 Normal Attack DMG.'
    )
    cost: Cost = Cost(any_dice_number = 3)

    def _attach_status(
        self, player_idx: int, charactor_idx: int
    ) -> List[CreateObjectAction]:
        """
        attach status
        """
        if (
            player_idx == self.position.player_idx
            and charactor_idx == self.position.charactor_idx
            and self.position.area == ObjectPositionType.CHARACTOR
        ):
            # equipped and switch to this charactor
            return [CreateObjectAction(
                object_position = self.position.set_area(
                    ObjectPositionType.CHARACTOR_STATUS),
                object_name = self.name,
                object_arguments = {}
            )]
        return []

    def event_handler_SWITCH_CHARACTOR(
        self, event: SwitchCharactorEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        If switch to this charactor, attach status.
        """
        return self._attach_status(event.action.player_idx, 
                                   event.action.charactor_idx)

    def event_handler_CHOOSE_CHARACTOR(
        self, event: ChooseCharactorEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        If choose this charactor, attach status.
        """
        return self._attach_status(event.action.player_idx, 
                                   event.action.charactor_idx)


class CapriciousVisage(SkillCostDecreaseArtifact):
    name: Literal['Capricious Visage']
    desc: str = (
        'When the character uses an Elemental Skill or equips a Talent: Spend '
        '1 less Elemental Die. (Once per Round)'
    )
    version: Literal['4.0'] = '4.0'
    cost: Cost = Cost(any_dice_number = 2)
    skill_label: int = CostLabels.ELEMENTAL_SKILL.value


class ShimenawasReminiscence(CapriciousVisage):
    name: Literal["Shimenawa's Reminiscence"]
    desc: str = (
        'When the character uses an Elemental Skill or equips a Talent: Spend '
        '1 less Elemental Die. (Once per Round) '
        "If the character has at least 2 Energy, this character's Normal "
        "Attacks and Elemental Skills will deal +1 DMG."
    )
    cost: Cost = Cost(any_dice_number = 3)

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL'],
    ) -> DamageIncreaseValue:
        """
        if this charactor has at least 2 energy and use normal attack or
        elemental skill, +1 damage.
        """
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, None
        ):
            # not current charactor using skill
            return value
        skill = match.get_object(value.position)
        if skill.skill_type not in [SkillType.NORMAL_ATTACK, 
                                    SkillType.ELEMENTAL_SKILL]:
            # skill type not match
            return value
        charactor = match.player_tables[
            self.position.player_idx].charactors[self.position.charactor_idx]
        if charactor.charge < 2:
            # not enough energy
            return value
        # increase damage
        assert mode == 'REAL'
        value.damage += 1
        return value


VermillionShimenawas = (
    ThunderingPoise | VermillionHereafter
    | CapriciousVisage | ShimenawasReminiscence 
)



from .venti import Venti, Stormeye, EmbraceOfWinds
from .maguu_kenki import (
    MaguuKenki, TranscendentAutomaton, 
    ShadowswordGallopingFrost, ShadowswordLoneGale
)
from .kaedehara_kazuha import (
    KaedeharaKazuha, PoeticsOfFuubutsu, AutumnWhirlwind
)
from .sucrose import Sucrose, LargeWindSpirit, ChaoticEntropy
from .jean import Jean, DandelionField, LandsOfDandelion
from .xiao import Xiao, ConquerorOfEvilGuardianYaksha
from .wanderer import Wanderer, GalesOfReverie


AnemoCharactors = (
    Sucrose | Jean | Venti | Xiao | KaedeharaKazuha | Wanderer
    # finally monsters
    | MaguuKenki
)
SummonsOfAnemoCharactors = (
    LargeWindSpirit | DandelionField | Stormeye | AutumnWhirlwind
    # finally monsters
    | ShadowswordGallopingFrost | ShadowswordLoneGale 
)
AnemoCharactorTalents = (
    ChaoticEntropy | LandsOfDandelion | EmbraceOfWinds 
    | ConquerorOfEvilGuardianYaksha | PoeticsOfFuubutsu | GalesOfReverie
    # finally monsters
    | TranscendentAutomaton
)



from typing import Any, List, Literal

from ...summon.base import AttackerSummonBase

from ...action import Actions
from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Summons


class Guoba(AttackerSummonBase):
    name: Literal['Guoba'] = 'Guoba'
    version: Literal['3.3'] = '3.3'
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.PYRO
    damage: int = 2


# Skills


class DoughFu(PhysicalNormalAttackBase):
    name: Literal['Dough-Fu'] = 'Dough-Fu'
    cost: Cost = PhysicalNormalAttackBase.get_cost(ElementType.PYRO)


class GuobaAttack(ElementalSkillBase):
    name: Literal['Guoba Attack'] = 'Guoba Attack'
    desc: str = '''Summons 1 Guoba.'''
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        ret: List[Actions] = []
        if self.is_talent_equipped(match):
            ret = super().get_actions(match)
        ret.append(self.create_summon('Guoba'))
        ret.append(self.charge_self(1))
        return ret


class Pyronado(ElementalBurstBase):
    name: Literal['Pyronado'] = 'Pyronado'
    desc: str = '''Deals 3 Pyro DMG, creates 1 Pyronado.'''
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 4,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_team_status('Pyronado')
        ]


# Talents


class Crossfire(SkillTalent):
    name: Literal['Crossfire']
    desc: str = (
        'Combat Action: When your active character is Xiangling, equip this '
        'card. After Xiangling equips this card, immediately use Guoba Attack '
        'once. When your Xiangling, who has this card equipped, uses Guoba '
        'Attack, she will also deal 1 Pyro DMG.'
    )
    version: Literal['3.3'] = '3.3'
    charactor_name: Literal['Xiangling'] = 'Xiangling'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 4
    )
    skill: GuobaAttack = GuobaAttack()


# charactor base


class Xiangling(CharactorBase):
    name: Literal['Xiangling']
    version: Literal['3.8'] = '3.8'
    desc: str = '''"Exquisite Delicacy" Xiangling'''
    element: ElementType = ElementType.PYRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        DoughFu | GuobaAttack | Pyronado
    ] = []
    faction: List[FactionType] = [
        FactionType.LIYUE
    ]
    weapon_type: WeaponType = WeaponType.POLEARM
    talent: Crossfire | None = None

    def _init_skills(self) -> None:
        self.skills = [
            DoughFu(),
            GuobaAttack(),
            Pyronado(),
        ]



from .favonius import FavoniusWeapons
from .skyward import SkywardWeapons
from .vanilla import VanillaWeapon
from .sacrificial import SacrificialWeapons
from .sumeru_forge_weapon import SumeruForgeWeapons
from .other_claymore import Claymores
from .other_catalyst import Catalysts
from .other_polearm import Polearms
from .other_bow import Bows
from .other_sword import Swords
from .old_version import OldVersionWeapons


Weapons = (
    VanillaWeapon | SacrificialWeapons | Catalysts | Bows | Claymores 
    | Polearms | Swords | SkywardWeapons | FavoniusWeapons | SumeruForgeWeapons
    # Finally old version weapons
    | OldVersionWeapons
)



from typing import Any, List, Literal, Tuple

from ...summon.base import AttackerSummonBase

from ...modifiable_values import CostValue, DamageValue
from ...event import GameStartEventArguments, RoundPrepareEventArguments

from ...action import (
    Actions, CreateObjectAction, MakeDamageAction
)
from ...struct import Cost, ObjectPosition

from ...consts import (
    CostLabels, DamageElementalType, DamageType, DieColor, ElementType, 
    FactionType, ObjectPositionType, PlayerActionLabels, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalNormalAttackBase, ElementalSkillBase, 
    PassiveSkillBase, CharactorBase, TalentBase
)


# Summons


class ChainsOfWardingThunder(AttackerSummonBase):
    name: Literal['Chains of Warding Thunder'] = 'Chains of Warding Thunder'
    desc: str = (
        'End Phase: Deal 1 Electro DMG. '
        'When this Summon is on the field: Your opponent will have to spend '
        '1 additional Elemental Die when performing "Switch Character." '
        '(Once per Round) Usage(s): 2'
    )
    version: Literal['3.7'] = '3.7'
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.ELECTRO
    damage: int = 1

    increase_cost_valid: bool = True

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        """
        When reaching prepare, reset usage
        """
        self.increase_cost_valid = True
        return []

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal['TEST', 'REAL']
    ) -> CostValue:
        """
        When increase_cost_valid, enemy perform charactor switch will increase
        1 cost.
        """
        if not self.increase_cost_valid:
            # out of usage, not modify
            return value
        if value.cost.label & CostLabels.SWITCH_CHARACTOR == 0:
            # not switch charactor, not modify
            return value
        if value.position.player_idx == self.position.player_idx:
            # self switch, not modify
            return value
        # add 1 cost
        value.cost.any_dice_number += 1
        if mode == 'REAL':
            self.increase_cost_valid = False
        return value


# Skills


class RockPaperScissorsCombo(ElementalSkillBase):
    name: Literal['Rock-Paper-Scissors Combo'] = 'Rock-Paper-Scissors Combo'
    desc: str = (
        'Deals 2 Electro DMG and then separately performs "Prepare Skill" for '
        'Rock-Paper-Scissors Combo: Scissors and '
        'Rock-Paper-Scissors Combo: Paper.'
    )
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 5
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create prepare skill
        """
        return super().get_actions(match) + [
            self.create_charactor_status('Rock-Paper-Scissors Combo: Scissors')
        ]


class RockPaperScissorsComboScissors(ElementalSkillBase):
    name: Literal[
        'Rock-Paper-Scissors Combo: Scissors'
    ] = 'Rock-Paper-Scissors Combo: Scissors'
    desc: str = (
        'Deals 2 Electro DMG and then performs "Prepare Skill" for '
        'Rock-Paper-Scissors Combo: Paper.'
    )
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost()

    def is_valid(self, match: Any) -> bool:
        """
        Cannot use by player directly
        """
        return False

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create prepare skill
        """
        return [
            self.attack_opposite_active(match, self.damage, self.damage_type),
            self.create_charactor_status('Rock-Paper-Scissors Combo: Paper')
        ]


class RockPaperScissorsComboPaper(ElementalSkillBase):
    name: Literal[
        'Rock-Paper-Scissors Combo: Paper'
    ] = 'Rock-Paper-Scissors Combo: Paper'
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost()

    def is_valid(self, match: Any) -> bool:
        """
        Cannot use by player directly
        """
        return False

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Do not charge.
        """
        return [
            self.attack_opposite_active(match, self.damage, self.damage_type),
        ]


class LightningLockdown(ElementalBurstBase):
    name: Literal['Lightning Lockdown'] = 'Lightning Lockdown'
    desc: str = '''Deals 2 Electro DMG, summons Chains of Warding Thunder.'''
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 3,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match) + [
            self.create_summon('Chains of Warding Thunder')
        ]


class ElectroCrystalCore(PassiveSkillBase):
    name: Literal['Electro Crystal Core'] = 'Electro Crystal Core'
    desc: str = (
        '(Passive) When the battle begins, this character gains '
        'Electro Crystal Core.'
    )

    def event_handler_GAME_START(
        self, event: GameStartEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        When game begin, gain electro crystal core
        """
        return [self.create_charactor_status(self.name)]


# Talents


class AbsorbingPrism(TalentBase):
    name: Literal['Absorbing Prism']
    desc: str = (
        'Combat Action: When your active character is Electro Hypostasis, '
        'heal that character for 3 HP and attach the Electro Crystal Core to '
        'them.'
    )
    version: Literal['3.7'] = '3.7'
    charactor_name: Literal['Electro Hypostasis'] = 'Electro Hypostasis'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 3
    )
    remove_when_used: bool = True

    def get_action_type(self, match: Any) -> Tuple[int, bool]:
        return PlayerActionLabels.CARD.value, True

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[MakeDamageAction | CreateObjectAction]:
        """
        Using this card will heal electro hypostasis by 3 hp and attach
        electro crystal core to it. No need to equip it.
        """
        charactor = match.player_tables[
            self.position.player_idx].get_active_charactor()
        assert charactor.name == self.charactor_name
        return [
            MakeDamageAction(
                source_player_idx = self.position.player_idx,
                target_player_idx = self.position.player_idx,
                damage_value_list = [
                    DamageValue(
                        position = self.position,
                        damage_type = DamageType.HEAL,
                        target_position = charactor.position,
                        damage = -3,
                        damage_elemental_type = DamageElementalType.HEAL,
                        cost = Cost(),
                    )
                ],
            ),
            CreateObjectAction(
                object_name = 'Electro Crystal Core',
                object_position = charactor.position.set_area(
                    ObjectPositionType.CHARACTOR_STATUS),
                object_arguments = {}
            )
        ]


# charactor base


class ElectroHypostasis(CharactorBase):
    name: Literal['Electro Hypostasis']
    version: Literal['3.7'] = '3.7'
    desc: str = '''Electro Hypostasis'''
    element: ElementType = ElementType.ELECTRO
    max_hp: int = 8
    max_charge: int = 2
    skills: List[
        ElementalNormalAttackBase | RockPaperScissorsCombo 
        | RockPaperScissorsComboScissors | RockPaperScissorsComboPaper
        | LightningLockdown | ElectroCrystalCore
    ] = []
    faction: List[FactionType] = [
        FactionType.MONSTER
    ]
    weapon_type: WeaponType = WeaponType.OTHER
    talent: None = None

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name = 'Electro Crystal Projection',
                damage_type = self.element,
                cost = ElementalNormalAttackBase.get_cost(self.element),
            ),
            RockPaperScissorsCombo(),
            RockPaperScissorsComboScissors(),
            RockPaperScissorsComboPaper(),
            LightningLockdown(),
            ElectroCrystalCore(),
        ]



from typing import Any, List, Literal

from ...event import RoundPrepareEventArguments, SkillEndEventArguments

from ...action import Actions, MakeDamageAction
from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, 
    ObjectPositionType, SkillType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Skills


class Frostgnaw(ElementalSkillBase):
    name: Literal['Frostgnaw'] = 'Frostgnaw'
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = ElementalSkillBase.get_cost(ElementType.CRYO)


class GlacialWaltz(ElementalBurstBase):
    name: Literal['Glacial Waltz'] = 'Glacial Waltz'
    desc: str = '''Deals 1 Cryo DMG, creates 1 Icicle.'''
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 4,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_team_status('Icicle')
        ]


# Talents


class ColdBloodedStrike(SkillTalent):
    name: Literal['Cold-Blooded Strike']
    desc: str = (
        'Combat Action: When your active character is Kaeya, equip this card. '
        'After Kaeya equips this card, immediately use Frostgnaw once. '
        'After your Kaeya, who has this card equipped, uses Frostgnaw, '
        'he heals himself for 2 HP. (Once per Round)'
    )
    version: Literal['3.3'] = '3.3'
    charactor_name: Literal['Kaeya'] = 'Kaeya'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 4,
    )
    skill: Frostgnaw = Frostgnaw()
    usage: int = 1

    def equip(self, match: Any) -> List[Actions]:
        """
        Reset usage
        """
        self.usage = 1
        return super().equip(match)

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        """
        Reset Usage
        """
        self.usage = 1
        return []

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        When attached charactor use elemental skill, then heal itself.
        """
        if not self.position.check_position_valid(
            event.action.position, match, player_idx_same = True,
            charactor_idx_same = True, target_area = ObjectPositionType.SKILL,
            source_area = ObjectPositionType.CHARACTOR,
        ):
            # not same player or charactor, or not skill, or not equipped
            return []
        if self.usage <= 0:
            # out of usage
            return []
        charactor = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx]
        skill = charactor.get_object(event.action.position)
        assert skill is not None
        if skill.skill_type != SkillType.ELEMENTAL_SKILL:
            # not elemental skill
            return []
        # heal itself
        self.usage -= 1
        return [self.skill.heal_self(match, 2)]


# charactor base


class Kaeya(CharactorBase):
    name: Literal['Kaeya']
    version: Literal['3.3'] = '3.3'
    desc: str = '''"Frostwind Swordsman" Kaeya'''
    element: ElementType = ElementType.CRYO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | Frostgnaw | GlacialWaltz
    ] = []
    faction: List[FactionType] = [
        FactionType.MONDSTADT
    ]
    weapon_type: WeaponType = WeaponType.SWORD
    talent: ColdBloodedStrike | None = None

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Ceremonial Bladework',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            Frostgnaw(),
            GlacialWaltz(),
        ]



from .companions import Companions
from .locations import Locations
from .items import Items
from .old_version import (
    OldVersionCompanions, OldVersionLocations, OldVersionItems
)


Supports = (
    Locations | Companions | Items
    # finally old version cards
    | OldVersionLocations | OldVersionCompanions | OldVersionItems
)



from src.lpsim.agents.interaction_agent import InteractionAgent
from src.lpsim.server.match import Match, MatchState
from src.lpsim.server.deck import Deck
from tests.utils_for_test import (
    check_hp, get_random_state, get_test_id_from_command, make_respond, 
    set_16_omni
)


def test_ayaka():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "skill 0 15 14 13",
            "TEST 2 p1c1 usage 1",
            "card 0 1 12 11",
            "card 0 0 10 9",
            "sw_char 1",
            "skill 1 8 7 6",
            "skill 0 5 4 3",
            "sw_char 0",
            "sw_char 1 2",
            "sw_char 0 1",
            "sw_char 1 0",
            "end",
            "skill 1 15 14 13",
            "sw_char 2 12",
            "TEST 1 10 3 8 4 6 8",
            "sw_char 1",
            "skill 2 11 10 9",
            "end",
            "sw_char 2 15",
            "sw_char 1",
            "sw_char 2 14",
            "TEST 1 10 3 4 4 0 8",
            "sw_char 0",
            "end",
            "skill 0 15 14 13"
        ],
        [
            "sw_card",
            "choose 2",
            "TEST 2 p0c0 usage 1",
            "TEST 1 10 10 10 10 10 8",
            "TEST 3 p1c2 app cryo",
            "sw_char 1 15",
            "skill 0 14 13 12",
            "sw_char 0 11",
            "TEST 1 10 8 10 7 6 8",
            "end",
            "skill 1 15 14 13",
            "TEST 1 10 5 10 4 6 8",
            "skill 0 12 11 10",
            "skill 0 9 8 7",
            "sw_char 1 6",
            "end",
            "choose 2",
            "sw_char 0 15",
            "sw_char 2 12",
            "sw_char 0 0",
            "skill 2 1 10 9",
            "end",
            "skill 0 15 14 13",
            "choose 2",
            "TEST 4 no card can use",
            "end"
        ]
    ]
    agent_0 = InteractionAgent(
        player_idx = 0,
        verbose_level = 0,
        commands = cmd_records[0],
        only_use_command = True
    )
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = cmd_records[1],
        only_use_command = True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state = get_random_state())
    # deck information
    deck = Deck.from_str(
        '''
        default_version:4.0
        charactor:Kamisato Ayaka
        charactor:Kamisato Ayaka
        charactor:Fischl
        Kanten Senmyou Blessing*30
        '''
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.charactor_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError('No need respond.')
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(' ')[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                cidx = int(cmd[2][3])
                usage = int(cmd[-1])
                status = match.player_tables[pidx].charactors[cidx].status
                assert len(status) == 1
                assert status[0].usage == usage
            elif test_id == 3:
                assert match.player_tables[1].charactors[
                    2].element_application == ['CRYO']
            elif test_id == 4:
                for req in match.requests:
                    assert req.name != 'UseCardRequest'
            else:
                raise AssertionError(f'Unknown test id {test_id}')
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == '__main__':
    test_ayaka()



"""
Team status that with old version. Note the version should not be initialized
by default to avoid using it accidently.
"""

from typing import Any, List, Literal

from .event_cards import FreshWindOfFreedom
from .base import ExtraAttackTeamStatus, UsageTeamStatus
from ...modifiable_values import DamageIncreaseValue
from ...consts import DamageElementalType, DamageType, IconType, SkillType
from ...event import AfterMakeDamageEventArguments
from ...action import RemoveObjectAction


class CatalyzingField_3_3(UsageTeamStatus):
    """
    Catalyzing field.
    """
    name: Literal['Catalyzing Field'] = 'Catalyzing Field'
    desc: str = (
        'When you deal Electro DMG or Pyro DMG to an opposing active '
        'charactor, DMG dealt +1.'
    )
    version: Literal['3.3']
    usage: int = 3
    max_usage: int = 3
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def value_modifier_DAMAGE_INCREASE(
            self, value: DamageIncreaseValue, match: Any,
            mode: Literal['TEST', 'REAL']) -> DamageIncreaseValue:
        """
        Increase damage for dendro or electro damages, and decrease usage.
        """
        if value.damage_type != DamageType.DAMAGE:
            # not damage, not modify
            return value
        if not self.position.check_position_valid(
            value.position, match, player_idx_same = True,
        ):
            # source not self, not activate
            return value
        if not self.position.check_position_valid(
            value.target_position, match, 
            player_idx_same = False, target_is_active_charactor = True,
        ):
            # target not enemy, or target not active charactor, not activate
            return value
        if value.damage_elemental_type in [
            DamageElementalType.DENDRO,
            DamageElementalType.ELECTRO,
        ] and self.usage > 0:
            value.damage += 1
            assert mode == 'REAL'
            self.usage -= 1
        return value

    def event_handler_MAKE_DAMAGE(
        self, event: AfterMakeDamageEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        When damage made, check whether the round status should be removed.
        Not trigger on AFTER_MAKE_DAMAGE because when damage made, run out
        of usage, but new one is generated, should remove first then generate
        new one, otherwise newly updated status will be removed.
        """
        return self.check_should_remove()


class RainbowBladework_3_3(UsageTeamStatus, ExtraAttackTeamStatus):
    name: Literal['Rainbow Bladework'] = 'Rainbow Bladework'
    desc: str = 'After your character uses a Normal Attack: Deal 1 Hydro DMG.'
    version: Literal['3.3']
    usage: int = 3
    max_usage: int = 3
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    trigger_skill_type: SkillType | None = SkillType.NORMAL_ATTACK
    damage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.HYDRO
    decrease_usage: bool = True


class WindAndFreedom_3_7(FreshWindOfFreedom):
    name: Literal['Wind and Freedom'] = 'Wind and Freedom'
    version: Literal['3.7']


OldVersionTeamStatus = (
    WindAndFreedom_3_7

    | CatalyzingField_3_3 | RainbowBladework_3_3
)



"""
Old version of charactors, summons and talents. Status are not defined here,
and old status will be implemented in their old_version.py.
"""
from .beidou_3_4 import Beidou_3_4
from .razor_3_3 import Razor_3_3
from .eula_3_5 import Eula_3_5, LightfallSword_3_5
from .sangonomiya_kokomi_3_5 import SangonomiyaKokomi_3_5
from .yoimiya_3_4 import Yoimiya_3_4
from .collei_3_3 import FloralSidewinder_3_3
from .maguu_kenki_3_3 import MaguuKenki_3_3
from .yoimiya_3_3 import Yoimiya_3_3
from .xingqiu_3_3 import Xingqiu_3_3
from .xingqiu_3_6 import Xingqiu_3_6
from .xiangling_3_3 import Xiangling_3_3
from .ganyu_3_3 import Ganyu_3_3, UndividedHeart_3_3
from .mirror_maiden_3_3 import MirrorMaiden_3_3
from .fatui_cryo_cicin_mage_3_7 import FatuiCryoCicinMage_3_7, CryoCicins_3_7
from .kamisato_ayato_3_6 import KamisatoAyato_3_6
from .tartaglia_3_7 import Tartaglia_3_7, AbyssalMayhemHydrospout_3_7
from .diona_3_3 import ShakenNotPurred_3_3


OldTalents = (
    AbyssalMayhemHydrospout_3_7

    | FloralSidewinder_3_3 | UndividedHeart_3_3 | ShakenNotPurred_3_3
)
OldSummons = (
    CryoCicins_3_7 

    | LightfallSword_3_5
)
OldCharactors = (
    Tartaglia_3_7 | FatuiCryoCicinMage_3_7

    | Xingqiu_3_6 | KamisatoAyato_3_6

    | Eula_3_5 | SangonomiyaKokomi_3_5 

    | Yoimiya_3_4 | Beidou_3_4

    | MaguuKenki_3_3 | Yoimiya_3_3 | Xingqiu_3_3 | Xiangling_3_3 | Ganyu_3_3
    | Razor_3_3 | MirrorMaiden_3_3
)



from typing import Any, List, Literal

from ...struct import Cost

from ...consts import (
    CostLabels, DamageElementalType, DamageType, IconType, ObjectPositionType, 
    SkillType
)

from ...modifiable_values import CostValue, DamageIncreaseValue, DamageValue

from ...action import MakeDamageAction, RemoveObjectAction, SkillEndAction

from ...event import MakeDamageEventArguments, RoundEndEventArguments
from .base import (
    DefendCharactorStatus, RoundCharactorStatus, UsageCharactorStatus
)


class Satiated(RoundCharactorStatus):
    name: Literal['Satiated'] = 'Satiated'
    desc: str = 'You cannot consume more Food this Round'
    version: Literal['3.3'] = '3.3'
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.FOOD] = IconType.FOOD


class JueyunGuoba(RoundCharactorStatus, UsageCharactorStatus):
    name: Literal['Jueyun Guoba'] = 'Jueyun Guoba'
    desc: str = (
        "During this Round, the target character's next Normal Attack "
        "deals +1 DMG."
    )
    version: Literal['3.3'] = '3.3'
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.ATK_UP] = IconType.ATK_UP

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        """
        When this charactor use Elemental Burst, add 3 damage.

        Although it seems allow increase background damage, but as no
        skill will damage only background, it will be sonsumed when foreground
        charactor receives damage, and we do not need to check it.
        """
        assert mode == 'REAL'
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            # not this charactor use elemental burst, not modify
            return value
        if self.usage <= 0:
            # no usage, not modify
            return value
        # elemental burst, modify
        value.damage += 1
        self.usage -= 1
        return value


class AdeptusTemptation(RoundCharactorStatus, UsageCharactorStatus):
    name: Literal["Adeptus' Temptation"] = "Adeptus' Temptation"
    desc: str = (
        "During this Round, the target character's next Elemental Burst "
        "deals +3 DMG."
    )
    version: Literal['3.3'] = '3.3'
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.ATK_UP] = IconType.ATK_UP

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        """
        When this charactor use Elemental Burst, add 3 damage.

        Although it seems allow increase background damage, but as no
        skill will damage only background, it will be sonsumed when foreground
        charactor receives damage, and we do not need to check it.
        """
        assert mode == 'REAL'
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, SkillType.ELEMENTAL_BURST
        ):
            # not this charactor use elemental burst, not modify
            return value
        if self.usage <= 0:
            # no usage, not modify
            return value
        # elemental burst, modify
        value.damage += 3
        self.usage -= 1
        return value


class LotusFlowerCrisp(DefendCharactorStatus, RoundCharactorStatus):
    name: Literal['Lotus Flower Crisp'] = 'Lotus Flower Crisp'
    desc: str = (
        "During this Round, the target character takes -3 DMG the next time."
    )
    version: Literal['3.3'] = '3.3'
    usage: int = 1
    max_usage: int = 1
    min_damage_to_trigger: int = 1
    max_in_one_time: int = 3


class NorthernSmokedChicken(RoundCharactorStatus, UsageCharactorStatus):
    name: Literal['Northern Smoked Chicken']
    desc: str = (
        "During this Round, the target character's next Normal Attack cost "
        "less 1 Unaligned Element."
    )
    version: Literal['3.3'] = '3.3'
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.BUFF] = IconType.BUFF

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal['TEST', 'REAL']
    ) -> CostValue:
        """
        If this charactor use normal attack, decrease one unaligned die cost.
        """
        if self.usage <= 0:  # pragma: no cover
            # no usage, not modify
            return value
        if not self.position.check_position_valid(
            value.position, match, player_idx_same = True, 
            charactor_idx_same = True, target_area = ObjectPositionType.SKILL,
        ):
            # not charactor use skill, not modify
            return value
        if value.cost.label & CostLabels.NORMAL_ATTACK.value == 0:
            # not normal attack, not modify
            return value
        # modify
        if value.cost.decrease_cost(None):
            # decrease success
            if mode == 'REAL':
                self.usage -= 1
        return value


class MushroomPizza(RoundCharactorStatus):
    name: Literal['Mushroom Pizza'] = 'Mushroom Pizza'
    desc: str = (
        'End Phase: Heal this charactor for 1 HP. '
        'Usage(s): 2'
    )
    version: Literal['3.3'] = '3.3'
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.HEAL] = IconType.HEAL

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        target_charactor = match.player_tables[
            self.position.player_idx].charactors[self.position.charactor_idx]
        return [MakeDamageAction(
            source_player_idx = self.position.player_idx,
            target_player_idx = self.position.player_idx,
            damage_value_list = [
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.HEAL,
                    target_position = target_charactor.position,
                    damage = -1,
                    damage_elemental_type = DamageElementalType.HEAL,
                    cost = Cost(),
                )
            ],
        )]


class MintyMeatRolls(RoundCharactorStatus):
    name: Literal['Minty Meat Rolls'] = 'Minty Meat Rolls'
    desc: str = (
        "During this Round, the target character's next 3 Normal Attacks cost "
        "less 1 Unaligned Element."
    )
    version: Literal['3.4'] = '3.4'
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.BUFF] = IconType.BUFF

    decrease_usage: int = 3

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal['TEST', 'REAL']
    ) -> CostValue:
        """
        If this charactor use normal attack, decrease one unaligned die cost.
        """
        if self.decrease_usage <= 0:  # pragma: no cover
            # no usage, not modify
            return value
        if not self.position.check_position_valid(
            value.position, match, player_idx_same = True, 
            charactor_idx_same = True, target_area = ObjectPositionType.SKILL,
        ):
            # not charactor use skill, not modify
            return value
        if value.cost.label & CostLabels.NORMAL_ATTACK.value == 0:
            # not normal attack, not modify
            return value
        # modify
        if value.cost.decrease_cost(None):
            # decrease success
            if mode == 'REAL':
                self.decrease_usage -= 1
        return value

    def event_handler_SKILL_END(
        self, event: SkillEndAction, match: Any
    ) -> List[RemoveObjectAction]:
        if self.decrease_usage == 0:
            self.usage = 0
        return self.check_should_remove()


class SashiMiPlatter(RoundCharactorStatus):
    name: Literal['Sashimi Platter'] = 'Sashimi Platter'
    desc: str = (
        "During this Round, the target character's Normal Attacks "
        "deals +1 DMG."
    )
    version: Literal['3.7'] = '3.7'
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.ATK_UP] = IconType.ATK_UP

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        """
        When this charactor use Elemental Burst, add 3 damage.

        Although it seems allow increase background damage, but as no
        skill will damage only background, it will be sonsumed when foreground
        charactor receives damage, and we do not need to check it.
        """
        assert mode == 'REAL'
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            # not this charactor use elemental burst, not modify
            return value
        if value.damage_from_element_reaction:
            # element reaction, not modify
            return value
        if value.damage_elemental_type == DamageElementalType.PIERCING:
            # piercing, not modify
            return value
        # elemental burst, modify
        value.damage += 1
        return value


class TandooriRoastChicken(RoundCharactorStatus):
    name: Literal['Tandoori Roast Chicken'] = 'Tandoori Roast Chicken'
    desc: str = (
        "During this Round, all your characters' next Elemental Skills "
        "deal +2 DMG."
    )
    version: Literal['3.7'] = '3.7'
    icon_type: Literal[IconType.ATK_UP] = IconType.ATK_UP

    usage: int = 1
    max_usage: int = 1

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        """
        When this charactor use Elemental Skill, add 2 damage.
        Logic same as AdeptusTemptation.
        """
        assert mode == 'REAL'
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, SkillType.ELEMENTAL_SKILL
        ):
            # not this charactor use elemental skill, not modify
            return value
        if self.usage <= 0:
            # no usage, not modify
            return value
        # elemental burst, modify
        value.damage += 2
        self.usage -= 1
        return value

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        return self.check_should_remove()


class ButterCrab(DefendCharactorStatus, RoundCharactorStatus):
    name: Literal['Butter Crab'] = 'Butter Crab'
    desc: str = (
        "During this Round, the target character takes -2 DMG the next time."
    )
    version: Literal['3.7'] = '3.7'
    usage: int = 1
    max_usage: int = 1
    min_damage_to_trigger: int = 1
    max_in_one_time: int = 2


FoodStatus = (
    Satiated | JueyunGuoba | AdeptusTemptation | LotusFlowerCrisp 
    | NorthernSmokedChicken | MushroomPizza | MintyMeatRolls | SashiMiPlatter
    | TandooriRoastChicken | ButterCrab
)



"""
Charactor status generated by system.
"""

from typing import Any, List, Literal
from .base import RoundCharactorStatus
from ...modifiable_values import DamageIncreaseValue
from ...consts import DamageElementalType, IconType
from ...event import MakeDamageEventArguments
from ...action import Actions


class Frozen(RoundCharactorStatus):
    """
    Frozen.
    """
    name: Literal['Frozen'] = 'Frozen'
    desc: str = (
        'Charactor cannot use skills. (Lasts until the end of this Round) '
        'When this charactor receives Pyro DMG or Physical DMG, '
        'removes this effect and increases DMG taken by 2.'
    )
    version: Literal['3.3'] = '3.3'
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.FROZEN] = IconType.FROZEN

    def value_modifier_DAMAGE_INCREASE(
            self, value: DamageIncreaseValue, match: Any,
            mode: Literal['TEST', 'REAL']) -> DamageIncreaseValue:
        """
        Increase damage for pyro and physical damages to self by 2, and 
        decrease usage.
        """
        if not value.is_corresponding_charactor_receive_damage(
            self.position, match
        ):
            # not this charactor receive damage, not modify
            return value
        if value.damage_elemental_type in [
            DamageElementalType.PYRO,
            DamageElementalType.PHYSICAL,
        ] and self.usage > 0:
            value.damage += 2
            assert mode == 'REAL'
            self.usage -= 1
        return value

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[Actions]:
        """
        When damage made, check whether the status should be removed.
        Not trigger on AFTER_MAKE_DAMAGE because when damage made, run out
        of usage, but new one is generated, should remove first then generate
        new one, otherwise newly updated status will be removed.
        """
        return list(self.check_should_remove())


SystemCharactorStatus = Frozen | Frozen



"""
Base class of charactors. Each charactor should have its own file, the file
contains the charactor class, talent class, talent card class, skill classes.
DO NOT implement status, summons, weapons, artifacts, etc. in this file, which
will break the import loop.
"""


from typing import List, Literal, Any, Tuple, get_origin, get_type_hints

from pydantic import validator

from ..card.equipment.artifact.base import ArtifactBase

from ..card.equipment.weapon.base import WeaponBase

from ..status.charactor_status.base import CharactorStatusBase

from ...utils import get_instance_from_type_unions

from ..event import MoveObjectEventArguments, UseSkillEventArguments
from ..consts import (
    ELEMENT_TO_DIE_COLOR, DamageElementalType, DamageType, ObjectType, 
    PlayerActionLabels, SkillType, WeaponType, ElementType, FactionType, 
    ObjectPositionType, CostLabels
)
from ..object_base import (
    ObjectBase, CardBase
)
from ..struct import Cost, DeckRestriction, ObjectPosition
from ..modifiable_values import DamageValue
from ..status import CharactorStatus
from ..card.equipment.artifact import Artifacts
from ..card.equipment.weapon import Weapons
from ..action import (
    ChargeAction, CreateObjectAction, MakeDamageAction, MoveObjectAction, 
    RemoveObjectAction, Actions, SkillEndAction, UseSkillAction
)


class SkillBase(ObjectBase):
    """
    Base class of skills.
    """
    name: str
    desc: str
    type: Literal[ObjectType.SKILL] = ObjectType.SKILL
    skill_type: SkillType
    damage_type: DamageElementalType
    damage: int
    cost: Cost
    cost_label: int
    position: ObjectPosition = ObjectPosition(
        player_idx = -1,
        area = ObjectPositionType.INVALID,
        id = -1,
    )

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        # set cost label into cost
        self.cost.label = self.cost_label
        # based on charactor id, calculate skill id.

    def get_action_type(self, match: Any) -> Tuple[int, bool]:
        """
        Get the action type of using the skill.

        Returns:
            Tuple[int, bool]: The first element is the action label, the second
                element is whether the action type is a combat action.
        """
        return PlayerActionLabels.SKILL.value, True

    def is_valid(self, match: Any) -> bool:
        """
        Check if the skill can be used.
        """
        return True

    def get_actions(self, match: Any) -> List[Actions]:
        """
        The skill is triggered, and get actions of the skill.
        By default, it will generate three actions:
        1. MakeDamageAction to attack the enemy active charactor with damage
            `self.damage` and damage type `self.damage_type`.
        2. ChargeAction to charge the active charactor by 1.
        """
        return [
            self.attack_opposite_active(match, self.damage, self.damage_type),
            self.charge_self(1),
        ]

    def event_handler_USE_SKILL(
        self, event: UseSkillEventArguments, match: Any
    ):
        """
        When use skill event triggered, check if this skill is used.
        If so, return self.get_actions()
        """
        if event.action.skill_position == self.position:
            return self.get_actions(match)
        return []

    # commonly used function for skills

    def is_talent_equipped(self, match: Any):
        charactor = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx]
        return charactor.talent is not None

    def create_team_status(
        self, name: str, args: Any = {}
    ) -> CreateObjectAction:
        position = ObjectPosition(
            player_idx = self.position.player_idx,
            area = ObjectPositionType.TEAM_STATUS,
            id = -1
        )
        return CreateObjectAction(
            object_name = name,
            object_position = position,
            object_arguments = args
        )

    def create_charactor_status(
        self, name: str, args: Any = {}
    ) -> CreateObjectAction:
        return CreateObjectAction(
            object_name = name,
            object_position = self.position.set_area(
                ObjectPositionType.CHARACTOR_STATUS),
            object_arguments = args
        )

    def create_summon(
        self, name: str, args: Any = {}
    ) -> CreateObjectAction:
        position = ObjectPosition(
            player_idx = self.position.player_idx,
            area = ObjectPositionType.SUMMON,
            id = -1
        )
        return CreateObjectAction(
            object_name = name,
            object_position = position,
            object_arguments = args
        )

    def charge_self(
        self, charge: int
    ) -> ChargeAction:
        return ChargeAction(
            player_idx = self.position.player_idx,
            charactor_idx = self.position.charactor_idx,
            charge = charge,
        )

    def attack_opposite_active(
        self, match: Any, damage: int, damage_type: DamageElementalType,
    ) -> MakeDamageAction:
        assert damage > 0
        target_table = match.player_tables[1 - self.position.player_idx]
        target_charactor_idx = target_table.active_charactor_idx
        target_charactor = target_table.charactors[target_charactor_idx]
        return MakeDamageAction(
            source_player_idx = self.position.player_idx,
            target_player_idx = 1 - self.position.player_idx,
            damage_value_list = [
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.DAMAGE,
                    target_position = target_charactor.position,
                    damage = damage,
                    damage_elemental_type = damage_type,
                    cost = self.cost.copy(),
                )
            ],
        )

    def heal_self(self, match: Any, heal: int) -> MakeDamageAction:
        charactor = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx]
        return MakeDamageAction(
            source_player_idx = self.position.player_idx,
            target_player_idx = self.position.player_idx,
            damage_value_list = [
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.HEAL,
                    target_position = charactor.position,
                    damage = - heal,
                    damage_elemental_type = DamageElementalType.HEAL,
                    cost = Cost(),
                )
            ],
        )

    def element_application_self(
        self, match: Any, element: DamageElementalType
    ) -> MakeDamageAction:
        charactor = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx]
        return MakeDamageAction(
            source_player_idx = self.position.player_idx,
            target_player_idx = self.position.player_idx,
            damage_value_list = [
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.ELEMENT_APPLICATION,
                    target_position = charactor.position,
                    damage = 0,
                    damage_elemental_type = element,
                    cost = self.cost.copy(),
                )
            ],
        )


class PhysicalNormalAttackBase(SkillBase):
    """
    Base class of physical normal attacks.
    """
    desc: str = """Deals 2 Physical DMG."""
    skill_type: Literal[SkillType.NORMAL_ATTACK] = SkillType.NORMAL_ATTACK
    damage_type: DamageElementalType = DamageElementalType.PHYSICAL
    damage: int = 2
    cost_label: int = CostLabels.NORMAL_ATTACK.value

    @staticmethod
    def get_cost(element: ElementType) -> Cost:
        return Cost(
            elemental_dice_color = ELEMENT_TO_DIE_COLOR[element],
            elemental_dice_number = 1,
            any_dice_number = 2,
        )


class ElementalNormalAttackBase(SkillBase):
    """
    Base class of elemental normal attacks.
    """
    desc: str = """Deals 1 _ELEMENT_ DMG."""
    skill_type: Literal[SkillType.NORMAL_ATTACK] = SkillType.NORMAL_ATTACK
    damage_type: DamageElementalType
    damage: int = 1
    cost_label: int = CostLabels.NORMAL_ATTACK.value

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        self.desc = self.desc.replace(
            '_ELEMENT_', self.damage_type.value.lower().capitalize())

    @staticmethod
    def get_cost(element: ElementType) -> Cost:
        return Cost(
            elemental_dice_color = ELEMENT_TO_DIE_COLOR[element],
            elemental_dice_number = 1,
            any_dice_number = 2,
        )


class ElementalSkillBase(SkillBase):
    """
    Base class of elemental skills.
    """
    desc: str = """Deals _DAMAGE_ _ELEMENT_ DMG."""
    skill_type: Literal[SkillType.ELEMENTAL_SKILL] = SkillType.ELEMENTAL_SKILL
    damage_type: DamageElementalType
    damage: int = 3
    cost_label: int = CostLabels.ELEMENTAL_SKILL.value

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        self.desc = self.desc.replace(
            '_ELEMENT_', self.damage_type.value.lower().capitalize()
        ).replace('_DAMAGE_', str(self.damage))

    @staticmethod
    def get_cost(element: ElementType) -> Cost:
        return Cost(
            elemental_dice_color = ELEMENT_TO_DIE_COLOR[element],
            elemental_dice_number = 3,
        )


class ElementalBurstBase(SkillBase):
    """
    Base class of elemental bursts.
    """
    desc: str = """Deals _DAMAGE_ _ELEMENT_ DMG."""
    skill_type: Literal[SkillType.ELEMENTAL_BURST] = SkillType.ELEMENTAL_BURST
    damage_type: DamageElementalType
    cost_label: int = CostLabels.ELEMENTAL_BURST.value

    @staticmethod
    def get_cost(element: ElementType, number: int, charge: int) -> Cost:
        return Cost(
            elemental_dice_color = ELEMENT_TO_DIE_COLOR[element],
            elemental_dice_number = number,
            charge = charge
        )

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        self.desc = self.desc.replace(
            '_ELEMENT_', self.damage_type.value.lower().capitalize())
        self.desc = self.desc.replace('_DAMAGE_', str(self.damage))

    def get_actions(self, match: Any) -> List[Actions]:
        """
        When using elemental burst, the charge of the active charactor will be
        reduced by `self.charge`.
        """
        return [
            self.charge_self(-self.cost.charge),
            self.attack_opposite_active(match, self.damage, self.damage_type),
        ]


class AOESkillBase(SkillBase):
    """
    Base class that deals AOE damage. Can inherit this class after previous
    classes to do AOE attack. It will attack active charactor damage+element, 
    back chractor back_damage_piercing.
    """
    desc: str = (
        'Deals _ACTIVE_ _ELEMENT_ DMG, deals _BACK_ Piercing DMG to all '
        'opposing characters on standby.'
    )
    back_damage: int

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        self.desc = self.desc.replace(
            '_ELEMENT_', self.damage_type.value.lower().capitalize())
        self.desc = self.desc.replace('_ACTIVE_', str(self.damage))
        self.desc = self.desc.replace('_BACK_', str(self.back_damage))

    def attack_opposite_active(
        self, match: Any, damage: int, damage_type: DamageElementalType,
    ) -> MakeDamageAction:
        """
        For AOE, modify attack opposite to also attack back charactors
        """
        target_table = match.player_tables[1 - self.position.player_idx]
        target_charactor_idx = target_table.active_charactor_idx
        target_charactor = target_table.charactors[target_charactor_idx]
        charactors = target_table.charactors
        damage_action = MakeDamageAction(
            source_player_idx = self.position.player_idx,
            target_player_idx = 1 - self.position.player_idx,
            damage_value_list = [
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.DAMAGE,
                    target_position = target_charactor.position,
                    damage = damage,
                    damage_elemental_type = damage_type,
                    cost = self.cost.copy(),
                )
            ],
        )
        for cid, charactor in enumerate(charactors):
            if cid == target_table.active_charactor_idx:
                continue
            if charactor.is_defeated:
                continue
            damage_action.damage_value_list.append(
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.DAMAGE,
                    target_position = charactor.position,
                    damage = self.back_damage,
                    damage_elemental_type = DamageElementalType.PIERCING,
                    cost = self.cost.copy(),
                )
            )
        return damage_action


class PassiveSkillBase(SkillBase):
    """
    Base class of passive skills.
    It has no cost and is always invalid (cannot be used).
    It has triggers to make effects.
    """
    skill_type: Literal[SkillType.PASSIVE] = SkillType.PASSIVE
    damage_type: DamageElementalType = DamageElementalType.PHYSICAL
    damage: int = 0
    cost: Cost = Cost()
    cost_label: int = 0

    def is_valid(self, match: Any) -> bool:
        """
        Passive skills are always invalid.
        """
        return False

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Passive skills are always invalid, so it will return an empty list.
        """
        raise AssertionError('Try to get actions of a passive skill')
        return []


class TalentBase(CardBase):
    """
    Base class of talents. Note almost all talents are skills, and will receive
    cost decrease from other objects.
    """
    name: str
    charactor_name: str
    type: Literal[ObjectType.TALENT] = ObjectType.TALENT
    cost_label: int = CostLabels.CARD.value | CostLabels.TALENT.value
    remove_when_used: bool = False

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        restriction = (
            f' (You must have {self.charactor_name} in your deck to add this '
            'card to your deck.)'
        )
        self.desc = self.desc.replace(restriction, '')
        self.desc += restriction

    def get_deck_restriction(self) -> DeckRestriction:
        """
        For talent cards, should contain the corresponding charactor.
        """
        return DeckRestriction(
            type = 'CHARACTOR', 
            name = self.charactor_name, 
            number = 1
        )

    def is_valid(self, match: Any) -> bool:
        """
        Only corresponding charactor is active charactor can equip this card.
        """
        if self.position.area != ObjectPositionType.HAND:
            # not in hand, cannot equip
            raise AssertionError('Talent is not in hand')
        table = match.player_tables[self.position.player_idx]
        return (table.charactors[table.active_charactor_idx].name
                == self.charactor_name)

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        """
        For most talent cards, can quip only on active charactor.
        """
        table = match.player_tables[self.position.player_idx]
        charactor = table.charactors[table.active_charactor_idx]
        return [charactor.position]

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[Actions]:
        """
        Act the talent. will place it into talent area.
        When other talent is equipped, remove the old one.
        For subclasses, inherit this and add other actions (e.g. trigger
        correcponding skills)
        """
        assert target is not None
        ret: List[Actions] = []
        table = match.player_tables[target.player_idx]
        charactor = table.charactors[target.charactor_idx]
        # check if need to remove current talent
        if charactor.talent is not None:
            ret.append(RemoveObjectAction(
                object_position = charactor.talent.position,
            ))
        new_position = charactor.position.set_id(self.position.id)
        ret.append(MoveObjectAction(
            object_position = self.position,
            target_position = new_position
        ))
        return ret

    def equip(self, match: Any) -> List[Actions]:
        return []

    def event_handler_MOVE_OBJECT(
        self, event: MoveObjectEventArguments, match: Any
    ) -> List[Actions]:
        """
        When this talent is moved from hand to charactor, it is considered
        as equipped, and will call `self.equip`.
        """
        if (
            event.action.object_position.id == self.id
            and event.action.object_position.area == ObjectPositionType.HAND
            and event.action.target_position.area 
            == ObjectPositionType.CHARACTOR
        ):
            # this talent equipped from hand to charactor
            return self.equip(match)
        return []


class SkillTalent(TalentBase):
    """
    Talents that trigger skills. They will get skill as input, which is
    saved as a private variable.
    """

    skill: SkillBase

    def is_valid(self, match: Any) -> bool:
        """
        Both TalentBase and SkillBase should be valid.
        """
        return super().is_valid(match) and self.skill.is_valid(match)

    def get_action_type(self, match: Any) -> Tuple[int, bool]:
        """
        For skill talent, the action label contains SKILL and is a combat 
        action.
        """
        return (
            PlayerActionLabels.CARD.value | PlayerActionLabels.SKILL.value,
            True
        )

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[Actions]:
        ret = super().get_actions(target, match)
        assert len(ret) > 0
        assert ret[-1].type == 'MOVE_OBJECT'
        target_position = ret[-1].target_position
        target_charactor = match.player_tables[
            target_position.player_idx
        ].charactors[target_position.charactor_idx]
        skills: List[SkillBase] = target_charactor.skills
        for s in skills:
            if s.name == self.skill.name:
                self.skill.position = s.position
                break
        else:
            raise AssertionError('Skill not found')
        ret.append(UseSkillAction(
            skill_position = self.skill.position
        ))
        # skill used, add SkillEndAction
        ret.append(SkillEndAction(
            position = self.skill.position,
            target_position = match.player_tables[
                1 - self.skill.position.player_idx
            ].get_active_charactor().position,
            skill_type = self.skill.skill_type,
        ))
        return ret


class CharactorBase(ObjectBase):
    """
    Base class of charactors.
    """
    name: str
    desc: str
    strict_version_validation: bool = False  # default accept higher versions
    version: str
    type: Literal[ObjectType.CHARACTOR] = ObjectType.CHARACTOR
    position: ObjectPosition = ObjectPosition(
        player_idx = -1,
        area = ObjectPositionType.INVALID,
        id = -1,
    )

    element: ElementType
    max_hp: int
    max_charge: int
    skills: List[SkillBase]

    # labels
    faction: List[FactionType]
    weapon_type: WeaponType

    # charactor status
    hp: int = -1
    charge: int = 0
    weapon: WeaponBase | None = None
    artifact: ArtifactBase | None = None
    talent: TalentBase | None = None
    status: List[CharactorStatusBase] = []
    element_application: List[ElementType] = []
    is_alive: bool = True

    @validator('status', each_item = True, pre = True)
    def parse_status(cls, v):
        return get_instance_from_type_unions(CharactorStatus, v)

    @validator('weapon', pre = True)
    def parse_weapon(cls, v):
        if v is None:
            return v
        return get_instance_from_type_unions(Weapons, v)

    @validator('artifact', pre = True)
    def parse_artifact(cls, v):
        if v is None:
            return v
        return get_instance_from_type_unions(Artifacts, v)

    @validator('version', pre = True)
    def accept_same_or_higher_version(cls, v: str, values):  # pragma: no cover
        msg = 'version annotation must be Literal with one str'
        if not isinstance(v, str):
            raise NotImplementedError(msg)
        version_hints = get_type_hints(cls)['version']
        if get_origin(version_hints) != Literal:
            raise NotImplementedError(msg)
        version_hints = version_hints.__args__
        if len(version_hints) > 1:
            raise NotImplementedError(msg)
        version_hint = version_hints[0]
        if values['strict_version_validation'] and v != version_hint:
            raise ValueError(
                f'version {v} is not equal to {version_hint}'
            )
        if v < version_hint:
            raise ValueError(
                f'version {v} is lower than {version_hint}'
            )
        return version_hint

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        if self.hp == -1:
            self.hp = self.max_hp
        old_skill_ids = [x.id for x in self.skills]
        self._init_skills()
        self.position = self.position  # set position to update skill positions
        if len(old_skill_ids):
            new_skill_ids = [x.id for x in self.skills]
            if len(new_skill_ids) != len(old_skill_ids):
                raise AssertionError('Skill number changed after init')
            for id, skill in zip(old_skill_ids, self.skills):
                skill.id = id
                skill.position = skill.position.set_id(id)

    def __setattr__(self, name: str, value: Any) -> None:
        """
        When position is edited, update skill positions.
        """
        super().__setattr__(name, value)
        if name == 'position':
            if self.position.player_idx < 0:
                # invalid position, do not update skill positions
                return
            for skill in self.skills:
                skill.position = ObjectPosition(
                    player_idx = self.position.player_idx,
                    charactor_idx = self.position.charactor_idx,
                    area = ObjectPositionType.SKILL,
                    id = skill.id,
                )

    def _init_skills(self) -> None:
        """
        Initialize skills. It will be called in __init__.
        """
        raise NotImplementedError

    @property
    def is_defeated(self) -> bool:
        return not self.is_alive

    @property
    def is_stunned(self) -> bool:
        """
        Check if the charactor is stunned.
        """
        stun_status_names = [
            'Frozen',
            'Petrification',
            'Mist Bubble',
            'Stun',
        ]
        for status in self.status:
            if status.name in stun_status_names:
                return True
        return False

    @property
    def damage_taken(self) -> int:
        """
        Get damage taken by the charactor.
        """
        return self.max_hp - self.hp

    def get_object_lists(self) -> List[ObjectBase]:
        """
        Get all objects of the charactor, order is passive skill, weapon, 
        artifact, talent, status. For status, order is their index in status 
        list, i.e. generated time.
        """
        result: List[ObjectBase] = [self]
        for skill in self.skills:
            result.append(skill)
        if self.weapon is not None:
            result.append(self.weapon)
        if self.artifact is not None:
            result.append(self.artifact)
        if self.talent is not None:
            result.append(self.talent)
        result += self.status
        return result

    def get_object(self, position: ObjectPosition) -> ObjectBase | None:
        """
        Get object by its position. If obect not exist, return None.
        """
        if position.area == ObjectPositionType.CHARACTOR_STATUS:
            for status in self.status:
                if status.id == position.id:
                    return status
            return None
        elif position.area == ObjectPositionType.SKILL:
            for skill in self.skills:
                if skill.id == position.id:
                    return skill
            raise AssertionError('Skill not found')
        else:
            assert position.area == ObjectPositionType.CHARACTOR
            if self.id == position.id:
                return self
            elif self.talent is not None and self.talent.id == position.id:
                return self.talent
            elif self.weapon is not None and self.weapon.id == position.id:
                return self.weapon
            elif self.artifact is not None and self.artifact.id == position.id:
                return self.artifact
            return None



from typing import Literal, List

from ..charactor_base import PhysicalNormalAttackBase
from ...consts import DieColor
from ...struct import Cost
from ..cryo.ganyu import Ganyu as G_3_7
from ..cryo.ganyu import UndividedHeart as UH_3_7
from ..cryo.ganyu import CelestialShower as CS_3_7
from ..cryo.ganyu import FrostflakeArrow, TrailOftheQilin


class CelestialShower(CS_3_7):
    desc: str = (
        'Deals 1 Cryo DMG, deals 1 Piercing DMG to all opposing characters on '
        'standby, summons 1 Sacred Cryo Pearl.'
    )
    damage: int = 1
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 3,
        charge = 2
    )


class UndividedHeart_3_3(UH_3_7):
    desc: str = (
        'Combat Action: When your active character is Ganyu, equip this card. '
        'After Ganyu equips this card, immediately use Frostflake Arrow once. '
        'When your Ganyu, who has this card equipped, uses Frostflake Arrow: '
        'Cryo DMG dealt by this Skill +1 '
        'if this Skill has been used before during this match, the Piercing '
        'DMG dealt to all opposing characters on standby is changed to 3.'
    )
    version: Literal['3.3'] = '3.3'


class Ganyu_3_3(G_3_7):
    version: Literal['3.3'] = '3.3'
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | TrailOftheQilin | FrostflakeArrow 
        | CelestialShower
    ] = []
    talent: UndividedHeart_3_3 | None

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Liutian Archery',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            TrailOftheQilin(),
            FrostflakeArrow(),
            CelestialShower()
        ]



from typing import Any, Literal, List

from ...event import CreateObjectEventArguments

from ...action import Actions, ChangeObjectUsageAction, CreateObjectAction

from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, 
    ObjectPositionType, WeaponType
)

from ..charactor_base import (
    CharactorBase, ElementalBurstBase, ElementalNormalAttackBase, 
    ElementalSkillBase, SkillTalent
)


class AllSchemesToKnow(ElementalSkillBase):
    name: Literal['All Schemes to Know'] = 'All Schemes to Know'
    desc: str = (
        'Deals 2 Dendro DMG, applies the Seed of Skandha to target character. '
        'If the target character already has Seed of Skandha applied to them, '
        'then apply Seed of Skandha to all opposing characters instead.'
    )
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.DENDRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.DENDRO,
        elemental_dice_number = 3,
    )

    def get_actions(self, match: Any) -> List[Actions]:
        ret: List[Actions] = []
        damage_ret = super().get_actions(match)
        # get target charactors and active charactor.
        player_idx = self.position.player_idx
        target_table = match.player_tables[1 - player_idx]
        target_charactors = target_table.charactors
        target_active_charactor = target_charactors[
            target_table.active_charactor_idx]
        # default set status to active charactor
        position = target_active_charactor.position.set_area(
            ObjectPositionType.CHARACTOR_STATUS)
        ret.append(
            CreateObjectAction(
                object_name = 'Seed of Skandha',
                object_position = position,
                object_arguments = {},
            )
        )
        # get seed of skandha status.
        active_charactor_status_names = [
            x.name for x in target_active_charactor.status
        ]
        if 'Seed of Skandha' in active_charactor_status_names:
            # apply to all opposing charactors.
            for charactor in target_charactors:
                if charactor.is_defeated:
                    continue
                if charactor.id != target_active_charactor.id:
                    position = charactor.position.set_area(
                        ObjectPositionType.CHARACTOR_STATUS)
                    ret.append(
                        CreateObjectAction(
                            object_name = 'Seed of Skandha',
                            object_position = position,
                            object_arguments = {},
                        )
                    )
        # apply damage after apply status.
        ret += damage_ret
        return ret


class AllSchemesToKnowTathata(ElementalSkillBase):
    name: Literal[
        'All Schemes to Know: Tathata'
    ] = 'All Schemes to Know: Tathata'
    desc: str = (
        'Deals 3 Dendro DMG, applies the Seed of Skandha to all opposing '
        'characters.'
    )
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.DENDRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.DENDRO,
        elemental_dice_number = 5,
    )

    def get_actions(self, match: Any) -> List[Actions]:
        ret: List[Actions] = []
        damage_ret = super().get_actions(match)
        # get target charactors and active charactor.
        player_idx = self.position.player_idx
        target_table = match.player_tables[1 - player_idx]
        target_charactors = target_table.charactors
        # apply to all opposing charactors.
        for charactor in target_charactors:
            if charactor.is_defeated:
                continue
            position = charactor.position.set_area(
                ObjectPositionType.CHARACTOR_STATUS)
            ret.append(
                CreateObjectAction(
                    object_name = 'Seed of Skandha',
                    object_position = position,
                    object_arguments = {},
                )
            )
        ret += damage_ret
        return ret


class IllusoryHeart(ElementalBurstBase):
    name: Literal['Illusory Heart'] = 'Illusory Heart'
    desc: str = '''Deals 4 Dendro DMG, creates 1 Shrine of Maya.'''
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.DENDRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.DENDRO,
        elemental_dice_number = 3,
        charge = 2,
    )

    def get_actions(self, match: Any) -> List[Actions]:
        ret = super().get_actions(match)
        ret.append(self.create_team_status('Shrine of Maya'))
        return ret


class TheSeedOfStoredKnowledge(SkillTalent):
    """
    Three effects are implemented in different places:
        Pyro: when Seed of Skandha is triggered, it will check whether need
            to add dendro damage.
        Electro: when Shrine of Maya is summoned, the status will check whether
            need to add one usage to Seed of Skandha.
        Hydro: when Shrine of Maya is summoned, this talent card will try
            to add one usage if is equipped.
    """
    name: Literal[
        'The Seed of Stored Knowledge'
    ] = 'The Seed of Stored Knowledge'
    charactor_name: Literal['Nahida'] = 'Nahida'
    desc: str = (
        'Combat Action: When your active character is Nahida, equip this '
        'card. After Nahida equips this card, immediately use Illusory Heart '
        'once. When your Nahida, who has this card equipped, is on the field, '
        'the following effects will take place based on your party\'s '
        'Elemental Types: Pyro: When the Shrine of Maya is on the field, '
        'opposing characters who trigger the Seed of Skandha due to Elemental '
        'Reactions they are affected by will have the Piercing DMG they take '
        'from the Seed of Skandha converted to Dendro DMG. '
        'Electro: When the Shrine of Maya enters the field, the Seed of '
        'Skandha currently present of the opposition will gain 1 Usage(s). '
        'Hydro: After your Nahida, who has this card equipped unleashes '
        'Shrine of Maya, Duration (Rounds) +1.'
    )
    version: Literal['3.7'] = '3.7'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.DENDRO,
        elemental_dice_number = 3,
        charge = 2,
    )
    skill: IllusoryHeart = IllusoryHeart()

    def event_handler_CREATE_OBJECT(
        self, event: CreateObjectEventArguments, match: Any
    ) -> List[ChangeObjectUsageAction]:
        """
        when Shrine of Maya is summoned, if this card is equipped, add one
        usage.
        """
        if event.action.object_name != 'Shrine of Maya':
            # not creating Shrine of Maya, do nothing.
            return []
        if not self.position.check_position_valid(
            event.action.object_position, match, player_idx_same = True,
            source_area = ObjectPositionType.CHARACTOR,
            source_is_active_charactor = True,
        ):
            # not self, or not equipped, or not active charactor, do nothing.
            return []
        table = match.player_tables[self.position.player_idx]
        has_hydro_charactor = False
        for charactor in table.charactors:
            if charactor.element == ElementType.HYDRO:
                has_hydro_charactor = True
                break
        if not has_hydro_charactor:
            # no hydro charactor, do nothing.
            return []
        team_status = table.team_status
        for status in team_status:
            if status.name == 'Shrine of Maya':
                # has Shrine Of Maya, add one usage.
                return [ChangeObjectUsageAction(
                    object_position = status.position,
                    change_usage = 1,
                    change_type = 'DELTA',
                )]
        raise AssertionError('Shrine of Maya not found.')


class Nahida(CharactorBase):
    name: Literal['Nahida']
    version: Literal['3.7'] = '3.7'
    desc: str = '''"Physic of Purity" Nahida'''
    element: ElementType = ElementType.DENDRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        ElementalNormalAttackBase | AllSchemesToKnow | AllSchemesToKnowTathata
        | IllusoryHeart
    ] = []
    faction: List[FactionType] = [
        FactionType.SUMERU,
    ]
    weapon_type: WeaponType = WeaponType.CATALYST
    talent: TheSeedOfStoredKnowledge | None = None

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name = 'Akara',
                damage_type = self.element,
                cost = ElementalNormalAttackBase.get_cost(self.element),
            ),
            AllSchemesToKnow(),
            AllSchemesToKnowTathata(),
            IllusoryHeart(),
        ]



from typing import Any, List, Literal

from ...modifiable_values import DamageIncreaseValue

from ...event import GameStartEventArguments

from ...summon.base import AttackerSummonBase

from ...action import Actions, ChargeAction, CreateObjectAction
from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, 
    SkillType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, PhysicalNormalAttackBase, 
    PassiveSkillBase, CharactorBase, SkillTalent
)


# Summons


class EyeOfStormyJudgment(AttackerSummonBase):
    name: Literal['Eye of Stormy Judgment'] = 'Eye of Stormy Judgment'
    desc: str = (
        'End Phase: Deal 1 Electro DMG. '
        "When this Summon is on the field: Your characters' Elemental Bursts "
        'deal +1 DMG.'
    )
    version: Literal['3.7'] = '3.7'
    usage: int = 3
    max_usage: int = 3
    damage_elemental_type: DamageElementalType = DamageElementalType.ELECTRO
    damage: int = 1

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        assert mode == 'REAL'
        if (
            not value.is_corresponding_charactor_use_damage_skill(
                self.position, match, SkillType.ELEMENTAL_BURST) 
            or value.damage_elemental_type == DamageElementalType.PIERCING
            or value.damage_from_element_reaction
        ):
            # not corresponding charactor use elemental burst, or is piercing
            # damage, or damage is from element reaction
            return value
        # add damage
        value.damage += 1
        return value


# Skills


class TranscendenceBalefulOmen(ElementalSkillBase):
    name: Literal[
        'Transcendence: Baleful Omen'] = 'Transcendence: Baleful Omen'
    desc: str = '''Summons 1 Eye of Stormy Judgment'''
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.PIERCING
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        create object
        """
        return [
            self.charge_self(1),
            self.create_summon('Eye of Stormy Judgment')
        ]


class SecretArtMusouShinsetsu(ElementalBurstBase):
    name: Literal[
        'Secret Art: Musou Shinsetsu'] = 'Secret Art: Musou Shinsetsu'
    desc: str = (
        'Deals 3 Electro DMG. All of your other characters gain 2 Energy.'
    )
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 4,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        ret = super().get_actions(match)
        table = match.player_tables[self.position.player_idx]
        for cid, charactor in enumerate(table.charactors):
            if charactor.is_alive and cid != self.position.charactor_idx:
                ret.append(ChargeAction(
                    player_idx = self.position.player_idx,
                    charactor_idx = cid,
                    charge = 2,
                ))
        return ret


class ChakraDesiderata(PassiveSkillBase):
    name: Literal['Chakra Desiderata'] = 'Chakra Desiderata'
    desc: str = (
        '(Passive) When the battle begins, this character gains '
        'Chakra Desiderata.'
    )

    def event_handler_GAME_START(
        self, event: GameStartEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        When game begin, gain stealth
        """
        return [self.create_charactor_status(self.name)]


# Talents


class WishesUnnumbered(SkillTalent):
    name: Literal['Wishes Unnumbered']
    desc: str = (
        'Combat Action: When your active character is Raiden Shogun, '
        'equip this card. After Raiden Shogun equips this card, immediately '
        'use Secret Art: Musou Shinsetsu once. When your Raiden Shogun, who '
        'has this card equipped, uses Secret Art: Musou Shinsetsu, it will '
        'deal +1 additional DMG for every point of Resolve consumed.'
    )
    version: Literal['3.7'] = '3.7'
    charactor_name: Literal['Raiden Shogun'] = 'Raiden Shogun'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 4,
        charge = 2
    )
    skill: SecretArtMusouShinsetsu = SecretArtMusouShinsetsu()


# charactor base


class RaidenShogun(CharactorBase):
    name: Literal['Raiden Shogun']
    version: Literal['3.7'] = '3.7'
    desc: str = '''"Plane of Euthymia" Raiden Shogun'''
    element: ElementType = ElementType.ELECTRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | TranscendenceBalefulOmen 
        | SecretArtMusouShinsetsu | ChakraDesiderata
    ] = []
    faction: List[FactionType] = [
        FactionType.INAZUMA
    ]
    weapon_type: WeaponType = WeaponType.POLEARM
    talent: WishesUnnumbered | None = None

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Origin',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            TranscendenceBalefulOmen(),
            SecretArtMusouShinsetsu(),
            ChakraDesiderata(),
        ]



from typing import Any, List, Literal

from ...event import RoundEndEventArguments

from ...summon.base import AttackerSummonBase

from ...action import Actions, CreateObjectAction, MakeDamageAction
from ...struct import Cost, ObjectPosition

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, 
    ObjectPositionType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Summons


class TenguJuuraiAmbush(AttackerSummonBase):
    name: Literal['Tengu Juurai: Ambush'] = 'Tengu Juurai: Ambush'
    desc: str = (
        'End Phase: Deal _DAMAGE_ Electro DMG, applies Crowfeather Cover to '
        'friendly active character.'
    )
    version: Literal['3.5'] = '3.5'
    usage: int = 1
    max_usage: int = 1
    damage_elemental_type: DamageElementalType = DamageElementalType.ELECTRO
    damage: int = 1

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction | CreateObjectAction]:
        ret: List[MakeDamageAction | CreateObjectAction] = []
        ret += super().event_handler_ROUND_END(event, match)
        active_idx = match.player_tables[
            self.position.player_idx].active_charactor_idx
        ret.append(CreateObjectAction(
            object_name = 'Crowfeather Cover',
            object_position = ObjectPosition(
                player_idx = self.position.player_idx,
                charactor_idx = active_idx,
                area = ObjectPositionType.CHARACTOR_STATUS,
                id = -1
            ),
            object_arguments = {}
        ))
        return ret


class TenguJuuraiStormcluster(TenguJuuraiAmbush):
    name: Literal['Tengu Juurai: Stormcluster'] = 'Tengu Juurai: Stormcluster'
    usage: int = 2
    max_usage: int = 2
    damage: int = 2


# Skills


class TenguStormcall(ElementalSkillBase):
    name: Literal['Tengu Stormcall'] = 'Tengu Stormcall'
    desc: str = '''Deals 1 Electro DMG, summons 1 Tengu Juurai: Ambush.'''
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match) + [
            self.create_summon('Tengu Juurai: Ambush')
        ]


class SubjugationKoukouSendou(ElementalBurstBase):
    name: Literal['Subjugation: Koukou Sendou'] = 'Subjugation: Koukou Sendou'
    desc: str = (
        'Deals 1 Electro DMG, summons 1 Tengu Juurai: Stormcluster.'
    )
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 4,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match) + [
            self.create_summon('Tengu Juurai: Stormcluster')
        ]


# Talents


class SinOfPride(SkillTalent):
    name: Literal['Sin of Pride']
    desc: str = (
        'Combat Action: When your active character is Kujou Sara, equip this '
        'card. After Kujou Sara equips this card, immediately use '
        'Subjugation: Koukou Sendou once. When Kujou Sara is active and has '
        'this card equipped, all allied Electro characters with Crowfeather '
        'Cover will deal +1 additional Elemental Skill and Elemental Burst '
        'DMG.'
    )
    version: Literal['3.5'] = '3.5'
    charactor_name: Literal['Kujou Sara'] = 'Kujou Sara'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 4,
        charge = 2
    )
    skill: SubjugationKoukouSendou = SubjugationKoukouSendou()


# charactor base


class KujouSara(CharactorBase):
    name: Literal['Kujou Sara']
    version: Literal['3.5'] = '3.5'
    desc: str = '''"Crowfeather Kaburaya" Kujou Sara'''
    element: ElementType = ElementType.ELECTRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | TenguStormcall | SubjugationKoukouSendou
    ] = []
    faction: List[FactionType] = [
        FactionType.INAZUMA
    ]
    weapon_type: WeaponType = WeaponType.BOW
    talent: SinOfPride | None = None

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Tengu Bowmanship',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            TenguStormcall(),
            SubjugationKoukouSendou()
        ]





from typing import Any, List, Literal

from ....action import CreateObjectAction, DrawCardAction

from ....struct import Cost
from .base import WeaponBase
from ....consts import ObjectPositionType, WeaponType


class FruitOfFulfillment(WeaponBase):
    name: Literal['Fruit of Fulfillment']
    desc: str = '''The character deals +1 DMG. When played: Draw 2 cards.'''
    cost: Cost = Cost(any_dice_number = 3)
    version: Literal['3.8'] = '3.8'
    weapon_type: WeaponType = WeaponType.CATALYST

    def equip(self, match: Any) -> List[DrawCardAction]:
        """
        draw 2 cards
        """
        return [DrawCardAction(
            player_idx = self.position.player_idx,
            number = 2,
            draw_if_filtered_not_enough = True
        )]


class KingsSquire(WeaponBase):
    name: Literal["King's Squire"]
    desc: str = (
        'The character deals +1 DMG. When played: The character to which this '
        'is attached will spend 2 less Elemental Dice next time they use an '
        'Elemental Skill or equip a Talent card.'
    )
    cost: Cost = Cost(same_dice_number = 3)
    version: Literal['4.0'] = '4.0'
    weapon_type: WeaponType = WeaponType.BOW

    def equip(self, match: Any) -> List[CreateObjectAction]:
        """
        attach status
        """
        return [CreateObjectAction(
            object_position = self.position.set_area(
                ObjectPositionType.CHARACTOR_STATUS),
            object_name = self.name,
            object_arguments = {}
        )]


class Moonpiercer(KingsSquire):
    name: Literal['Moonpiercer']
    version: Literal['4.1'] = '4.1'
    weapon_type: WeaponType = WeaponType.POLEARM


SumeruForgeWeapons = FruitOfFulfillment | KingsSquire | Moonpiercer



from typing import Any, List, Literal

from ...summon.base import SummonBase

from ...modifiable_values import DamageValue
from ...event import RoundEndEventArguments

from ...action import (
    Actions, ChangeObjectUsageAction, MakeDamageAction, RemoveObjectAction
)
from ...struct import Cost

from ...consts import (
    DamageElementalType, DamageType, DieColor, ElementType, FactionType, 
    IconType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Charactor status. DO NOT define here, define in server/status/characor_status
# Here is just example.


# Summons


class LightfallSword(SummonBase):
    name: Literal['Lightfall Sword'] = 'Lightfall Sword'
    desc: str = (
        'When Eula uses a Normal Attack or Elemental Skill, this card will '
        'accumulate 2 Zeal stacks, but Eula will not gain Energy. '
        'End Phase: Discard this card and deal _DAMAGE_ Physical DMG. '
        'Each Zeal stack adds 1 DMG to this damage instance. '
        "(Effects on this card's Usage will apply to Zeal.)"
    )
    version: Literal['3.8'] = '3.8'
    usage: int = 0
    max_usage: int = 999
    damage: int = 3
    damage_elemental_type: DamageElementalType = DamageElementalType.PHYSICAL
    icon_type: Literal[IconType.COUNTER] = IconType.COUNTER

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.desc = self.desc.replace('_DAMAGE_', str(self.damage))

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction | RemoveObjectAction]:
        """
        When round end, make damage to the opponent and remove self
        """
        player_idx = self.position.player_idx
        target_table = match.player_tables[1 - player_idx]
        target_charactor = target_table.get_active_charactor()
        return [
            MakeDamageAction(
                source_player_idx = player_idx,
                target_player_idx = 1 - player_idx,
                damage_value_list = [
                    DamageValue(
                        position = self.position,
                        damage_type = DamageType.DAMAGE,
                        target_position = target_charactor.position,
                        damage = self.damage + self.usage,
                        damage_elemental_type = self.damage_elemental_type,
                        cost = Cost(),
                    )
                ],
            ),
            RemoveObjectAction(
                object_position = self.position,
            )
        ]


# Skills


class FavoniusBladeworkEdel(PhysicalNormalAttackBase):
    name: Literal['Favonius Bladework - Edel'] = 'Favonius Bladework - Edel'
    cost: Cost = PhysicalNormalAttackBase.get_cost(ElementType.CRYO)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        If summon Lightfall Sword exists, not charge but add two usage.
        """
        summons = match.player_tables[self.position.player_idx].summons
        LS_idx = -1
        for sidx, summon in enumerate(summons):
            if summon.name == 'Lightfall Sword':
                LS_idx = sidx
                break
        ret: List[Actions] = []
        ret.append(self.attack_opposite_active(
            match, self.damage, self.damage_type))
        if LS_idx != -1:
            # add LS usage
            ret.append(ChangeObjectUsageAction(
                object_position = summons[LS_idx].position,
                change_type = 'DELTA',
                change_usage = 2,
            ))
        else:
            # charge
            ret.append(self.charge_self(1))
        return ret


class IcetideVortex(ElementalSkillBase):
    name: Literal['Icetide Vortex'] = 'Icetide Vortex'
    desc: str = (
        'Deals 2 Cryo DMG. If this character has not yet gained Grimheart, '
        'they will gain Grimheart.'
    )
    version: Literal['3.8'] = '3.8'
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        If summon Lightfall Sword exists, not charge but add two or three 
        usage, based on whether has talent.
        """
        summons = match.player_tables[self.position.player_idx].summons
        LS_idx = -1
        for sidx, summon in enumerate(summons):
            if summon.name == 'Lightfall Sword':
                LS_idx = sidx
                break
        ret: List[Actions] = []
        ret.append(self.attack_opposite_active(
            match, self.damage, self.damage_type))
        if LS_idx != -1:
            # add LS usage
            usage = 2
            if self.is_talent_equipped(match):
                # has talent, add one more
                usage = 3
            ret.append(ChangeObjectUsageAction(
                object_position = summons[LS_idx].position,
                change_type = 'DELTA',
                change_usage = usage,
            ))
        else:
            # charge
            ret.append(self.charge_self(1))
        status = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx].status
        for s in status:
            if s.name == 'Grimheart':
                # has Grimheart, return
                return ret
        # no Grimheart, add Grimheart
        ret.append(self.create_charactor_status(
            'Grimheart', { 'version': self.version }))
        return ret


class GlacialIllumination(ElementalBurstBase):
    name: Literal['Glacial Illumination'] = 'Glacial Illumination'
    desc: str = '''Deals 2 Cryo DMG, summons 1 Lightfall Sword.'''
    version: Literal['3.8'] = '3.8'
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 3,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and summon
        """
        return super().get_actions(match) + [
            self.create_summon('Lightfall Sword', { 'version': self.version })
        ]


# Talents


class WellspringOfWarLust(SkillTalent):
    name: Literal['Wellspring of War-Lust']
    desc: str = (
        'Combat Action: When your active character is Eula, equip this card. '
        'After Eula equips this card, immediately use Glacial Illumination '
        'once. After your Eula, who has this card equipped, uses Icetide '
        'Vortex, this will generate 1 more Zeal for Lightfall Sword.'
    )
    version: Literal['3.5'] = '3.5'
    charactor_name: Literal['Eula'] = 'Eula'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 3,
        charge = 2
    )
    skill: GlacialIllumination = GlacialIllumination()


# charactor base


class Eula(CharactorBase):
    name: Literal['Eula']
    version: Literal['3.8'] = '3.8'
    desc: str = '''"Dance of the Shimmering Wave" Eula'''
    element: ElementType = ElementType.CRYO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        FavoniusBladeworkEdel | IcetideVortex | GlacialIllumination
    ] = []
    faction: List[FactionType] = [
        FactionType.MONDSTADT
    ]
    weapon_type: WeaponType = WeaponType.CLAYMORE
    talent: WellspringOfWarLust | None = None

    def _init_skills(self) -> None:
        self.skills = [
            FavoniusBladeworkEdel(),
            IcetideVortex(),
            GlacialIllumination()
        ]



"""
Base classes of objects in the game table. They are all subclasses of
ObjectBase. CardBase are defined here because like ObjectBase, they are used 
in many other places. Other objects are defined in their own files.
"""


import time
import random

from .event import UseCardEventArguments
from ..utils import BaseModel
from typing import List, Literal, Any, Tuple, get_origin, get_type_hints
from pydantic import validator
from .action import Actions, ActionTypes, RemoveCardAction
from .consts import (
    ObjectType, ObjectPositionType, CostLabels, PlayerActionLabels
)
from .modifiable_values import ModifiableValueTypes
from .struct import (
    DeckRestriction, MultipleObjectPosition, ObjectPosition, Cost
)


used_object_ids = set()
ID_MOD_NUM = 86400
ID_MULTI_NUM = 1000000
ID_RAND_NUM = 1024


class ObjectBase(BaseModel):
    """
    Base class of objects in the game table. All objects in the game table 
    should inherit from this class.
    """
    type: ObjectType = ObjectType.EMPTY
    position: ObjectPosition
    id: int = -1

    # If the object is in trashbin, event handlers except those in this list
    # will not available.
    available_handler_in_trashbin: List[ActionTypes] = []

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        # check event handler name valid
        for k in dir(self):
            if k[:14] == 'event_handler_':
                k = k[14:]
                try:
                    k = ActionTypes(k)
                except ValueError:
                    raise ValueError(f'Invalid event handler name: {k}')
            if k[:15] == 'value_modifier_':
                k = k[15:]
                try:
                    k = ModifiableValueTypes(k)
                except ValueError:
                    raise ValueError(f'Invalid value modifier name: {k}')
        # if id is -1, generate a new id
        if self.id == -1:
            self.renew_id()
        used_object_ids.add(self.id)
        # set id in position
        self.position = self.position.set_id(self.id)

    def renew_id(self):
        """
        Renew the id of the object.
        """
        while True:
            new_id = (
                int(time.time() % ID_MOD_NUM * ID_MULTI_NUM) 
                * ID_RAND_NUM + random.randint(0, ID_RAND_NUM - 1)
            )
            if new_id in used_object_ids:  # pragma: no cover
                continue
            self.id = new_id
            used_object_ids.add(self.id)
            self.position = self.position.set_id(self.id)
            break


class CardBase(ObjectBase):
    """
    Base class of all real cards. 
    """
    name: str
    desc: str
    type: Literal[ObjectType.CARD, ObjectType.WEAPON, ObjectType.ARTIFACT,
                  ObjectType.TALENT, ObjectType.SUPPORT] = ObjectType.CARD
    strict_version_validation: bool = False  # default accept higher versions
    version: str
    position: ObjectPosition = ObjectPosition(
        player_idx = -1,
        area = ObjectPositionType.INVALID,
        id = -1,
    )
    cost: Cost
    cost_label: int = CostLabels.CARD.value
    remove_when_used: bool = True
    target_position_type_multiple: Literal[False] = False

    @validator('version', pre = True)
    def accept_same_or_higher_version(cls, v: str, values):  # pragma: no cover
        msg = 'version annotation must be Literal with one str'
        if not isinstance(v, str):
            raise NotImplementedError(msg)
        version_hints = get_type_hints(cls)['version']
        if get_origin(version_hints) != Literal:
            raise NotImplementedError(msg)
        version_hints = version_hints.__args__
        if len(version_hints) > 1:
            raise NotImplementedError(msg)
        version_hint = version_hints[0]
        if values['strict_version_validation'] and v != version_hint:
            raise ValueError(
                f'version {v} is not equal to {version_hint}'
            )
        if v < version_hint:
            raise ValueError(
                f'version {v} is lower than {version_hint}'
            )
        return version_hint

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        # set cost label into cost
        self.cost.label = self.cost_label
        assert self.cost.original_value is None

    def get_deck_restriction(self) -> DeckRestriction:
        """
        Get the deck restriction of the card. It will be checked when deck is
        created.
        """
        return DeckRestriction(type = 'NONE', name = '', number = 0)

    def get_action_type(self, match: Any) -> Tuple[int, bool]:
        """
        Get the action type of using the card. 

        Returns:
            Tuple[int, bool]: The first element is the action label, the second
                element is whether the action type is a combat action.
        """
        return PlayerActionLabels.CARD.value, False

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        """
        Get the targets of the card.
        """
        raise NotImplementedError()

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[Actions]:
        """
        Act the card. It will return a list of actions.

        Arguments:
            target (ObjectPosition | None): The target of the action.
                for cards that do not need to specify target, target is None.
                Note: the ID of the target may not be the same as the ID of the
                card, e.g. equipping artifact. Reconstruct correct 
                ObjectPosition if needed.
            match (Any): The match object.
        """
        raise NotImplementedError()

    def is_valid(self, match: Any) -> bool:
        """
        Check if the card can be used. Note that this function will not check
        the cost of the card.
        """
        return True

    def event_handler_USE_CARD(
        self, event: UseCardEventArguments, match: Any
    ) -> List[Actions]:
        """
        If this card is used, trigger events
        """
        actions: List[Actions] = []
        if event.action.card_position.id != self.id:
            # not this card
            return actions
        if self.remove_when_used:
            actions.append(RemoveCardAction(
                position = ObjectPosition(
                    player_idx = self.position.player_idx,
                    area = ObjectPositionType.HAND,
                    id = self.id,
                ),
                remove_type = 'USED',
            ))
        # target is None, otherwise should match class
        assert (
            event.action.target is None
            or (
                isinstance(event.action.target, MultipleObjectPosition)
                == self.target_position_type_multiple
            )
        )
        actions += self.get_actions(
            target = event.action.target,  # type: ignore
            match = match,
        )
        return actions


class MultiTargetCardBase(CardBase):
    """
    Base class of cards that can target multiple targets.
    """
    target_position_type_multiple: Literal[True] = True

    def get_targets(self, match: Any) -> List[MultipleObjectPosition]:
        raise NotImplementedError()

    def get_actions(
        self, target: MultipleObjectPosition | None, match: Any
    ) -> List[Actions]:
        """
        Act the card. It will return a list of actions.

        Arguments:
            target (MultipleObjectPosition | None): The target of the action.
                for cards that do not need to specify target, target is None.
                Note: the ID of the target may not be the same as the ID of the
                card, e.g. equipping artifact. Reconstruct correct 
                ObjectPosition if needed.
            match (Any): The match object.
        """
        raise NotImplementedError()


CardBases = CardBase | MultiTargetCardBase



from typing import List, Literal

from ..charactor_base import PhysicalNormalAttackBase
from ...consts import DieColor
from ...struct import Cost
from ..electro.razor import Razor as R_3_8
from ..electro.razor import LightningFang as LF_3_8
from ..electro.razor import ClawAndThunder


class LightningFang(LF_3_8):
    desc: str = (
        'Deals 5 Electro DMG. This character gains The Wolf Within.'
    )
    damage: int = 5
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 3,
        charge = 3
    )


class Razor_3_3(R_3_8):
    version: Literal['3.3']
    skills: List[
        PhysicalNormalAttackBase | ClawAndThunder | LightningFang
    ] = []
    max_charge: int = 3

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Steel Fang',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            ClawAndThunder(),
            LightningFang()
        ]



from typing import Any, List, Literal

from ...event import RoundEndEventArguments

from ...summon.base import AttackerSummonBase

from ...modifiable_values import DamageValue

from ...action import ActionTypes, Actions, MakeDamageAction
from ...struct import Cost

from ...consts import (
    DamageElementalType, DamageType, DieColor, ElementType, FactionType, 
    WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalNormalAttackBase, ElementalSkillBase, 
    CharactorBase, SkillTalent
)


# Summons


class BakeKurage(AttackerSummonBase):
    name: Literal['Bake-Kurage'] = 'Bake-Kurage'
    desc: str = (
        'End Phase: Deal 1 Hydro DMG, heal your active character for 1 HP.'
    )
    version: Literal['3.5'] = '3.5'
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.HYDRO
    damage: int = 1

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        ret = super().event_handler_ROUND_END(event, match)
        active_charactor = match.player_tables[
            self.position.player_idx].get_active_charactor()
        make_damage_action = ret[-1]
        assert make_damage_action.type == ActionTypes.MAKE_DAMAGE
        make_damage_action.damage_value_list.append(DamageValue(
            position = self.position,
            damage_type = DamageType.HEAL,
            target_position = active_charactor.position,
            damage = -1,
            damage_elemental_type = DamageElementalType.HEAL,
            cost = Cost(),
        ))
        # if our Kokomi with talent and Ceremonial Garment, increase damage
        charactors = match.player_tables[self.position.player_idx].charactors
        is_kokomi_talent_status: bool = False
        for c in charactors:
            if (
                c.name == 'Sangonomiya Kokomi'
                and c.talent is not None
            ):
                # our Kokomi with talent
                status_found = False
                for s in c.status:
                    if s.name == 'Ceremonial Garment':
                        # Ceremonial Garment exists
                        status_found = True
                        break
                if status_found:
                    # Ceremonial Garment exists
                    is_kokomi_talent_status = True
                    break
        if is_kokomi_talent_status:
            # increase damage
            make_damage_action.damage_value_list[0].damage += 1
        return ret


# Skills


class KuragesOath(ElementalSkillBase):
    name: Literal["Kurage's Oath"] = "Kurage's Oath"
    desc: str = (
        'This character gains Hydro Application and summons 1 Bake-Kurage.'
    )
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.PIERCING
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        application and create object
        """
        return [
            self.charge_self(1),
            self.element_application_self(match, DamageElementalType.HYDRO),
            self.create_summon('Bake-Kurage'),
        ]


class NereidsAscension(ElementalBurstBase):
    name: Literal["Nereid's Ascension"] = "Nereid's Ascension"
    desc: str = (
        'Deals 2 Hydro DMG. Heals all allied characters for 1 point. '
        'This character gains Ceremonial Garment.'
    )
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 3,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        damage, heal and create charactor status
        """
        ret = super().get_actions(match)
        action = MakeDamageAction(
            source_player_idx = self.position.player_idx,
            target_player_idx = self.position.player_idx,
            damage_value_list = []
        )
        charactors = match.player_tables[self.position.player_idx].charactors
        for charactor in charactors:
            if charactor.is_alive:
                action.damage_value_list.append(DamageValue(
                    position = self.position,
                    damage_type = DamageType.HEAL,
                    target_position = charactor.position,
                    damage = -1,
                    damage_elemental_type = DamageElementalType.HEAL,
                    cost = Cost(),
                ))
        ret.append(action)
        ret.append(self.create_charactor_status('Ceremonial Garment'))
        # if has talent, re-create Bake-Kurage
        if self.is_talent_equipped(match):
            ret.append(self.create_summon('Bake-Kurage'))
        return ret


# Talents


class TamakushiCasket(SkillTalent):
    name: Literal['Tamakushi Casket']
    desc: str = (
        'Combat Action: When your active character is Sangonomiya Kokomi, '
        'equip this card. After Sangonomiya Kokomi equips this card, '
        "immediately use Nereid's Ascension once. When your Sangonomiya "
        "Kokomi, who has this card equipped, uses Nereid's Ascension: If "
        "Bake-Kurage is on the field, its Usage(s) will be refreshed. While "
        'Ceremonial Garment exists, Bake-Kurage deals +1 DMG'
    )
    version: Literal['3.5'] = '3.5'
    charactor_name: Literal['Sangonomiya Kokomi'] = 'Sangonomiya Kokomi'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 3,
        charge = 2
    )
    skill: NereidsAscension = NereidsAscension()


# charactor base


class SangonomiyaKokomi(CharactorBase):
    name: Literal['Sangonomiya Kokomi']
    version: Literal['3.6'] = '3.6'
    desc: str = '''"Pearl of Wisdom" Sangonomiya Kokomi'''
    element: ElementType = ElementType.HYDRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        ElementalNormalAttackBase | KuragesOath | NereidsAscension
    ] = []
    faction: List[FactionType] = [
        FactionType.INAZUMA
    ]
    weapon_type: WeaponType = WeaponType.CATALYST
    talent: TamakushiCasket | None = None

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name = 'The Shape of Water',
                damage_type = self.element,
                cost = ElementalNormalAttackBase.get_cost(self.element),
            ),
            KuragesOath(),
            NereidsAscension()
        ]



from typing import Literal
from .base import AttackerSummonBase
from ..consts import DamageElementalType


class BurningFlame(AttackerSummonBase):
    name: Literal['Burning Flame'] = 'Burning Flame'
    desc: str = '''End Phase: Deal 1 Pyro DMG. (Can stack. Max 2 stacks.)'''
    version: Literal['3.3'] = '3.3'
    usage: int = 1
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.PYRO
    damage: int = 1
    renew_type: Literal['ADD'] = 'ADD'



from typing import Literal

from ...consts import DieColor
from ...struct import Cost
from ..dendro.collei import FloralSidewinder as FloralSidewinder_3_4


class FloralSidewinder_3_3(FloralSidewinder_3_4):
    version: Literal['3.3']
    cost: Cost = Cost(
        elemental_dice_color = DieColor.DENDRO,
        elemental_dice_number = 3
    )



"""
Arcane legend cards.
"""

from typing import Any, List, Literal

from ...modifiable_values import DamageValue

from ...action import (
    ConsumeArcaneLegendAction, CreateDiceAction, CreateObjectAction, 
    MakeDamageAction
)

from ...struct import Cost, ObjectPosition
from ...consts import (
    ELEMENT_TO_DAMAGE_TYPE, CostLabels, DamageType, ElementType, 
    ObjectPositionType, ObjectType
)
from ...object_base import CardBase


class ArcaneLegendBase(CardBase):
    type: Literal[ObjectType.ARCANE] = ObjectType.ARCANE
    cost_label: int = CostLabels.CARD.value | CostLabels.ARCANE.value

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[ConsumeArcaneLegendAction]:
        """
        Consume arcane legend.
        """
        assert target is None
        return [ConsumeArcaneLegendAction(
            player_idx = self.position.player_idx
        )]


class AncientCourtyard(ArcaneLegendBase):
    name: Literal['Ancient Courtyard']
    desc: str = (
        'You must have a character who has already equipped a Weapon or '
        'Artifact: The next time you play a Weapon or Artifact card in this '
        'Round: Spend 2 less Elemental Dice.'
    )
    version: Literal['3.8'] = '3.8'
    cost: Cost = Cost(arcane_legend = True)

    def is_valid(self, match: Any) -> bool:
        """
        You must have a character who has already equipped a Weapon or Artifact
        """
        for charactor in match.player_tables[
                self.position.player_idx].charactors:
            if charactor.weapon is not None or charactor.artifact is not None:
                return True
        return False

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        """
        No targets.
        """
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[ConsumeArcaneLegendAction | CreateObjectAction]:
        """
        create team status
        """
        assert target is None
        ret: List[ConsumeArcaneLegendAction | CreateObjectAction] = []
        ret += super().get_actions(target, match)
        ret.append(CreateObjectAction(
            object_name = self.name,
            object_position = ObjectPosition(
                player_idx = self.position.player_idx,
                area = ObjectType.TEAM_STATUS,
                id = 0,
            ),
            object_arguments = {}
        ))
        return ret


class CovenantOfRock(ArcaneLegendBase):
    name: Literal['Covenant of Rock']
    desc: str = (
        'Can only be played when you have 0 Elemental Dice left: '
        'Generate 2 different Elemental Dice.'
    )
    version: Literal['3.8'] = '3.8'
    cost: Cost = Cost(arcane_legend = True)

    def is_valid(self, match: Any) -> bool:
        """
        Can only be played when you have 0 Elemental Dice left.
        """
        return len(match.player_tables[
            self.position.player_idx].dice.colors) == 0

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        """
        No targets.
        """
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[ConsumeArcaneLegendAction | CreateDiceAction]:
        """
        Generate 2 different Elemental Dice.
        """
        assert target is None
        ret: List[ConsumeArcaneLegendAction | CreateDiceAction] = []
        ret += super().get_actions(target, match)
        ret.append(CreateDiceAction(
            player_idx = self.position.player_idx,
            number = 2,
            different = True
        ))
        return ret


class JoyousCelebration(ArcaneLegendBase):
    name: Literal['Joyous Celebration']
    desc: str = (
        'Your active character must be one of the following elemental types '
        'to play this card: Cryo/Hydro/Pyro/Electro/Dendro: The element '
        'corresponding to your active character\'s Elemental Type will be '
        'applied to all your characters.'
    )
    version: Literal['3.8'] = '3.8'
    cost: Cost = Cost(arcane_legend = True)

    def is_valid(self, match: Any) -> bool:
        """
        Your active character must be one of the following elemental types to
        play this card: Cryo/Hydro/Pyro/Electro/Dendro
        """
        charactor = match.player_tables[
            self.position.player_idx].get_active_charactor()
        return charactor.element in [
            ElementType.CRYO, ElementType.HYDRO, ElementType.PYRO,
            ElementType.ELECTRO, ElementType.DENDRO
        ]

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        """
        No targets.
        """
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[ConsumeArcaneLegendAction | MakeDamageAction]:
        """
        apply element to all your characters.
        """
        assert target is None
        ret: List[ConsumeArcaneLegendAction | MakeDamageAction] = []
        ret += super().get_actions(target, match)
        damage_action = MakeDamageAction(
            source_player_idx = self.position.player_idx,
            target_player_idx = self.position.player_idx,
            damage_value_list = [
            ],
        )
        active_charactor = match.player_tables[
            self.position.player_idx].get_active_charactor()
        element = active_charactor.element
        for c in match.player_tables[self.position.player_idx].charactors:
            if c.is_defeated:
                continue
            damage_action.damage_value_list.append(
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.ELEMENT_APPLICATION,
                    target_position = c.position,
                    damage = 0,
                    damage_elemental_type = ELEMENT_TO_DAMAGE_TYPE[element],
                    cost = self.cost.copy(),
                )
            )
        return ret + [damage_action]


class FreshWindOfFreedom(ArcaneLegendBase):
    name: Literal['Fresh Wind of Freedom']
    desc: str = (
        'In this Round, when an opposing character is defeated during your '
        'Action, you can continue to act again when that Action ends.'
    )
    version: Literal['4.1'] = '4.1'
    cost: Cost = Cost(arcane_legend = True)

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[ConsumeArcaneLegendAction | CreateObjectAction]:
        """
        Create Wind and Freedom team status.
        """
        assert target is None
        return super().get_actions(target, match) + [
            CreateObjectAction(
                object_name = self.name,
                object_position = ObjectPosition(
                    player_idx = self.position.player_idx,
                    area = ObjectPositionType.TEAM_STATUS,
                    id = -1,
                ),
                object_arguments = {},
            )
        ]


ArcaneLegendCards = (
    AncientCourtyard | CovenantOfRock | JoyousCelebration
    | FreshWindOfFreedom
)
